(* ::Package:: *)

(* 
Author: Andreas Mueller
Last updated: 29 June 2019
Purpose: Implements the recursive relations for derivatives of kinematic mapping and Jacobian minors presented in [1],[2], and [3]
References: 
[1] A. Mueller: Higher-Order Analysis of Kinematic Singularities of Lower Pair Linkages and Serial Manipulators, ASME J. Mech. Rob., 2017, DOI: 10.1115/1.4038528
[2] A. Mueller: Local Kinematic Analysis of Closed-Loop Linkages -Mobility, Singularities, and Shakiness, ASME J. Mech. Rob., Vol. 8, August 2016, 041013-1, DOI: 10.1115/1.4032778
[3] A. Mueller: An Overview of Formulae for the Higher-Order Kinematics of Lower-Pair Chains wit Applications in Robotics and Mechanism Theorym submitted to Mech. Mach. Theory
*)

BeginPackage["HigherOrderKinematics`","Combinatorica`"];

(****  If you are using Mathematica version older than version 9, you will need to replace the above by

BeginPackage["HigherOrderKinematics`",{"Combinatorica`","VectorAnalysis`"}];

Then also CrossProduct must be used below.
*)

TwistTose3::usage="ad.";
SE3ad::usage = "Computes the Lie bracket of two screw coordinate vectors X,Y in se(3).";
ScrewVector::usage = "Determines the screw coordinate vector from the position vector, orientation vector, and pitch.";

SE3Exp::usage = "Exp mapping of screw coordinate vector.";
SE3AdMatrix::usage = "Returns the matrix for the Adjoint operation.";

f::usage = "f[YList,q] returns the kinematic mapping";

St::usage = "Determines the mapping S for given q(t).";
Sxyz::usage = "Determines the mapping S(q,x,y,z,...) for given x,y,z,...";
Sx::usage = "Determines the mapping S for given x";

dSdt::usage = "Determines the k-th time derivative of the mapping S.";
DS::usage = "Determines the k-th derivative of the mapping S.";
DSJoint::usage= "Determine the k-th derivative of the joint screw coordinate vector S.";

df::usage = "Determine the k-th differential of the kinematic mapping f.";
dfinv::usage = "Determine the k-th differential of the inverse matrix as value of kinematic mapping f.";
h::usage = "Determine the k-th differential of the mapping h.";
dSJoint::usage = "Determine the k-th differential dS(x) of the joint screw coordinate vector S.";

DmList::usage = "Construct a list of i-th time derivative of all k-minors of the constraint Jacobian.\\This works for a lingle loop.";
Dm::usage = "Determines the time derivative of the submatrix according to multiindices a and b.\\This works for a lingle loop.";
DmListMultiLoop::usage = "Construct a list of i-th time derivative of all k-minors of the constraint Jacobian.\\This works for multiloop linkages.";
DmMultiLoop::usage = "Determines the time derivative of the submatrix according to multiindices a and b.\\This works for multiloop linkages.";

Dqdt::usage = "The k-th (time) derivative of the solution of loop closure constraints.";

SE3InvolutiveClosure::usage = "SE3InvolutiveClosure[{X1,...,Xn}] computes the involutive closure of {X1,...,Xn}.";

ZeroVector::usage = "Vector consisting of n zeros.";

(* Begin private section of the package *)
Begin["`Private`"];

(**********************************************************************************************************
          Elementary functions on screws/se(3)
 **********************************************************************************************************)
(* construction of screw coordinate vector from position vector p, orientation unit vector e, and pitch h *)
Unprotect[ScrewVector];
ScrewVector[e_, p_, h__: 0] := 
 Module[{v,r}, If[e === {0, 0, 0}, v = p, v = -Cross[e, p] + h*e];
  r = Join[e, v]
  ]
Protect[ScrewVector];

(* construction of se(3) matrix from twist vector *)
Unprotect[TwistTose3];
TwistTose3[X_] := Module[{w = Take[X,3]},
Return[
ArrayFlatten[{{{{0,-w[[3]],w[[2]]},{w[[3]],0,-w[[1]]},{-w[[2]],w[[1]],0}},
Transpose[{Take[X,-3]}]},{{{0,0,0}},{{0}}}}]
  ]
];
Protect[TwistTose3];

(* Lie bracket of two screw coordinate vectors X and Y *)
Unprotect[SE3ad];
SE3ad[X_, Y_] := 
 Join[Cross[Take[X, {1, 3}], Take[Y, {1, 3}]], 
  Cross[Take[X, {1, 3}], Take[Y, {4, 6}]] - 
   Cross[Take[Y, {1, 3}], Take[X, {4, 6}]]]
Protect[SE3ad];

(**********************************************************************************************************
          Exp mapping of screws.
 **********************************************************************************************************)
Unprotect[SE3Exp];
SE3Exp[X_,t_]:=Module[{xi=Take[X,{1,3}],eta=Take[X,{4,6}],R,p,w},
If[xi==={0,0,0}, 
	R=IdentityMatrix[3];p=eta*t,
	w={{0,-xi[[3]],xi[[2]]},
	   {xi[[3]],0,-xi[[1]]},
	   {-xi[[2]],xi[[1]],0}};
   R=IdentityMatrix[3]+Sin[t] w+(1-Cos[t]) w.w;
   p=(IdentityMatrix[3]-R).(w.eta)+xi (xi.eta)*t;
];
Return[ArrayFlatten[{{R,Transpose[{p}]},{{{0,0,0}},{{1}}}}]]
];
Protect[SE3Exp];

Unprotect[SE3AdMatrix];
SE3AdMatrix[C_]:=Module[{R,p},
R=Take[C,{1,3},{1,3}];
p={C[[1,4]],C[[2,4]],C[[3,4]]};
Return[
	ArrayFlatten[{{R,{{0,0,0},{0,0,0},{0,0,0}}},{{{0,-p[[3]],p[[2]]},{p[[3]],0,-p[[1]]},{-p[[2]],p[[1]],0}}.R,R}}]]];
Protect[SE3AdMatrix];

Unprotect[f];
f[YList_,q_]:=Module[{C=IdentityMatrix[4],n=Length[q],i},
If[Length[YList]!=n,Return[]];
For[i=1,i<=n,i++,C=C.SE3Exp[YList[[i]],q[[i]]]];
Return[C];
];
Protect[f];


(**********************************************************************************************************
          Recursive expressions for the time derivatives of the constraint mapping
          q[t] must be a function of time. The results are in terms of time derivatives of q[t].
 **********************************************************************************************************)
(* Determine the mapping S defined in [1] equation (8) given q (t) *)
(* TODO: "t" as argument !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*)
Unprotect[St];
St[ i_, q_, SList_] := Module[{j, n = Length[SList]},
   If[i > n, Return[]];
   Sum[SList[[j]]*D[q[[j]], {Global`t, 1}], {j, 1, i}]
   ];
Protect[St];

(* Determine the k-th time derivative of the mapping S.
   This implements equation (9) in [1], respectively equation (11) in [2] *)
Unprotect[dSdt];
dSdt[i_, q_, SList_, k_] := Module[{j, n = Length[SList]},
  If[i > n, Return[]];
  If[k == 0, Return[St[i, q, SList]]];
  Return[Sum[
       Sum[
        Binomial[k, l]*dSJointdt[j, q, SList, k - l]*
         D[q[[j]], {Global`t, l+1}], {l, 0, k}], {j, 1, i}]
         ];
   ];
Protect[dSdt];
 
(* Determine the k-th time derivative of the joint screw coordinate vector S_i. 
   This implements equation (10) in [1], respectively equation (12) in [2] *)
Unprotect[dSJointdt];
dSJointdt[i_, q_, SList_, k_] := Module[{j, l, n = Length[SList]},
   If[i > n, Return[]];
   If[k == 0, Return[SList[[i]]]];
   If[k == 1, Return[SE3ad[St[i, q, SList], SList[[i]]]]];
   If[k > 1, Return[Sum [
      Binomial[k - 1, l]*
       SE3ad[dSdt[i, q, SList, l], dSJointdt[i, q, SList, k - l - 1]], {l,
        0, k - 1}]
     ]
    ]];
Protect[dSJointdt];

(**********************************************************************************************************
          Recursive expressions for the derivatives of the constraint mapping
          xyz is a list of vectors x,y,z,... representing the variables for 1st, 2nd, 3rd, ... order
 **********************************************************************************************************)
(* Determine the mapping S defined in [1] equation (8) for given variables x,y,z,... *)
Unprotect[Sxyz];
Sxyz[i_,xyz_,SList_] := Module[{j, n = Length[SList]},
	If[i > n, Return[]];
	Sum[SList[[j]]*xyz[[1,j]],{j,1,i}]
	];
Protect[Sxyz];

(* Determine the k-th derivative of the mapping S.
   This implements equation (9) in [1], respectively equation (11) in [2] *)
Unprotect[DS];
DS[i_,xyz_,SList_,k_] := Module[{j, n = Length[SList]},
	If[i > n, Return[]];
	If[k == 0, Return[Sxyz[i,xyz,SList]]];
	Return[Sum[
	     Sum[
		  Binomial[k,l]*DSJoint[j,xyz,SList,k-l]*xyz[[1+l,j]],{l,0,k}],{j,1,i}]
		  ];
	];
Protect[DS];

(* Determine the k-th derivative of the joint screw coordinate vector S_i. 
   This implements equation (10) in [1], respectively equation (12) in [2] *)
Unprotect[DSJoint];
DSJoint[i_, xyz_, SList_, k_] := Module[{j, l, n = Length[SList]},
   If[i > n, Return[]];
   If[k == 0, Return[SList[[i]]]];
   If[k == 1, Return[SE3ad[Sxyz[i, xyz, SList], SList[[i]]]]];
   If[k > 1, Return[Sum [
      Binomial[k - 1, l]*
       SE3ad[DS[i, xyz, SList, l], 
        DSJoint[i, xyz, SList, k - l - 1]], {l, 0, k - 1}]
     ]
    ]];
Protect[DSJoint];

(**********************************************************************************************************
       Recursive expressions for the differential of the kinematic mapping
 **********************************************************************************************************)
Unprotect[Sx];
Sx[i_,x_,SList_] := Module[{j, n = Length[SList]},
	If[i > n, Return[]];
	Sum[SList[[j]]*x[[j]],{j,1,i}]
	];
Protect[Sx];

Unprotect[dSJoint];
dSJoint[i_,x_,SList_,k_] := Module[{j,r,n = Length[SList]},
	If[i > n, Return[]];
	If[k == 0, Return[SList[[i]]]];
	If[k >= 1, Return[Sum[
	     Sum[Binomial[k-1,r]*SE3ad[dSJoint[j,x,SList,r],dSJoint[i,x,SList,k-r-1]]*x[[j]],
	     {r,0,k-1}],{j,1,i}]];
	];
];
Protect[dSJoint];

Unprotect[h];
h[x_,SList_,k_] := Module[{i,n = Length[SList]},
Return[TwistTose3[Sum[x[[i]]*dSJoint[i,x,SList,k-1],{i,1,n}]]]];
Protect[h];

Unprotect[df];
df[x_,SList_,k_] := Module[{i,n = Length[SList]},
	If[k == 0, Return[IdentityMatrix[4]]];
	If[k == 1, Return[TwistTose3[Transpose[SList].x]]];
	If[k > 1, Return[ h[x,SList,k]-
Sum[Binomial[k-1,i-1]*df[x,SList,i].dfinv[x,SList,k-i],{i,1,k-1}]];
	];
];
Protect[df];

Unprotect[dfinv];
dfinv[x_,SList_,k_] := Module[{i,n = Length[SList]},
	If[k == 0, Return[IdentityMatrix[4]]];
	If[k ==1, Return[-TwistTose3[Transpose[SList].x]]];
If[k > 1, Return[
-Sum[Binomial[k,i]*df[x,SList,i].dfinv[x,SList,k-i],{i,1,k}]];
	];
];
Protect[dfinv];


(**********************************************************************************************************
       Recursive expressions for the time derivatives / differential
       of the minors of the constraint Jacobian for a single loop
 **********************************************************************************************************)
(* Determine the time derivative of the submatrix according to multiindices a and b. 
   This implements equation (28) in appendix of [1] 
   Input: 
         i - Order of the time derivative to be determined
         a, b - Multiindeces that select the minor
         SList - List with the joint screw coordiantes
         dSList - List of lists of time derivatives / differentials of joint screw coordinates up to the order=Length[dSList] *)
Dm[i_, a_, b_, SList_, dSList_, Planar_: False] := 
  Module[{pi, p, nList, muList, piList, j, k, l, m = 0, Mat, 
    MatReduced},
   If[Length[a] != Length[b], Return[]];
   piList = Partitions[i];
   For[k = 1, k <= Length[piList], k++,
    pi = piList[[k]];
    p = Length[pi];
    nList = {};
    For[j = 1, j <= i, j++, nList = Append[nList, Count[pi, j]]];
    muList = Permutations[b, {p}];
    For[j = 1, j <= Length[muList], j++,
     Mat = SubMat[pi, muList[[j]], b, SList, dSList, Planar];
     MatReduced = Table[{}, {l, Length[a]}];
     For[l = 1, l <= Length[a], l++,
      MatReduced[[l]] = Mat[[a[[l]]]];
      ];
     m = m + 
       Det[MatReduced]*i!/(Product[pi[[l]]!, {l, p}]*Product[nList[[l]]!, {l, i}]);
     ];
    ];
   Return[m];
   ];

(* Construct a list of i-th time derivative / differentials of all k-minors of the constraint Jacobian *)
DmList[i_, k_, SList_, dSList_, Planar_: False] := 
 Module[{a, b, l, j, nJoints},
  nJoints = Length[SList];
  If[Planar == True, a = Subsets[Range[3], {k}], 
   a = Subsets[Range[6], {k}]];
  b = Subsets[Range[nJoints], {k}];
  Flatten[Table[
    Dm[i, a[[l]], b[[j]], SList, dSList, Planar], {l, 
     Length[a]}, {j, Length[b]}]]
  ]

(**********************************************************************************************************
       Recursive expressions for the time derivatives / differential
       of the minors of the constraint Jacobian for a multiloop linkage
 **********************************************************************************************************)
DmMultiLoop[i_, a_, b_, SLists_, dSLists_, Planar_: False] := 
  Module[{nLoops, pi, p, nList, muList, piList, j, k, l, m = 0, Mat, 
    MatReduced},
   If[Length[a] != Length[b], Return[]];
   nLoops = Length[SLists];
   piList = Partitions[i];
   For[k = 1, k <= Length[piList], k++,
    pi = piList[[k]];
    p = Length[pi];
    nList = {};
    For[j = 1, j <= i, j++, nList = Append[nList, Count[pi, j]]];
    muList = Permutations[b, {p}];
    For[j = 1, j <= Length[muList], j++,
     Mat = 
      ArrayFlatten[
       Table[{SubMat[pi, muList[[j]], b, SLists[[l]], dSLists[[l]], 
          Planar]}, {l, nLoops}]];
     MatReduced = Table[{}, {l, Length[a]}];
     For[l = 1, l <= Length[a], l++,
      MatReduced[[l]] = Mat[[a[[l]]]];
      ];
     m = m + 
       Det[MatReduced]*
        i!/(Product[pi[[l]]!, {l, p}]*Product[nList[[l]]!, {l, i}]);
     ];
    ];
   Return[m];
   ];

DmListMultiLoop[i_, k_, SLists_, dSLists_, Planar_: False] := 
 Module[{a, b, l, j, nJoints, nLoops},
  nLoops = Length[SLists];
  nJoints = Length[SLists[[1]]];
  If[Planar == True, a = Subsets[Range[3*nLoops], {k}], 
   a = Subsets[Range[6*nLoops], {k}]];
  b = Subsets[Range[nJoints], {k}];
  Flatten[Table[
    DmMultiLoop[i, a[[l]], b[[j]], SLists, dSLists, Planar], {l, 
     Length[a]}, {j, Length[b]}]]
  ]
  
(* For given pi and mu, extract the submatrix according to the multiindex b. *)
SubMat[pi_, mu_, b_, SList_, dSList_, Planar_: False] := Module[{p, l, J, J2, Mat, se2Proj},
   se2Proj = {{0, 0, 1, 0, 0, 0}, {0, 0, 0, 1, 0, 0}, {0, 0, 0, 0, 1, 0}}; (* selection matrix for se (2) in standard representation *)
   p = Length[pi];
   J = SList;
   For[l = 1, l <= p, l++,
    J[[mu[[l]]]] = dSList[[pi[[l]], mu[[l]]]];
    ];
   J2 = Table[{}, {l, 1, Length[b]}];
   For[l = 1, l <= Length[b], l++,
    J2[[l]] = J[[b[[l]]]];
    ];
   If[Planar == True, Return[se2Proj.Transpose[J2]],
    Return[Transpose[J2]]];
   ];


Unprotect[Dqdt];
(* DuList - list of time derivatives of the vector of joint independent joint variables
   SList - list of joint screw coordinates
   Degree - degree up to which the derivatives of the dependent joints are computed
   Id - Index set of depend coordinates in increasing order
   Iu - Index set of independ coordinates in increasing order
   P - Selection matrix extracting the relevant components from the constraints
*)
Dqdt[DuList_, SList_, Degree_, P_, Id_, Iu_, qMap_] := 
  Module[{n, nu, nd, DqList, DSJointList, tmp, Jd, Ju, JdInv,
    DJdList, DJdInvList, DDList, F, DFList, DSList, DSJointk, DJuList, DJuk, DJdk, DJdInvk,
    k, i, j, l, kq, DSk, SHIFT},
	SHIFT = 1;
   	n = Length[SList];
   	nu = Length[Iu]; nd = Length[Id];
   	DSJointList = {SList};
   	tmp = Transpose[Table[SList[[Iu[[j]]]], {j, 1, nu}]];
   	Ju = Table[tmp[[P[[j]]]], {j, 1, Length[P]}];
   	DJuList = {Ju}; 
   	tmp = Transpose[Table[SList[[Id[[j]]]], {j, 1, nd}]];
   	Jd = Table[tmp[[P[[j]]]], {j, 1, Length[P]}]; 
   	JdInv = Inverse[Jd]; 
   	DJdList = {Jd};
   	DJdInvList = {JdInv};
   	DDList = {-JdInv.Ju};
   	F = ArrayFlatten[{{-JdInv.Ju}, {IdentityMatrix[nu]}}];
   	DFList = {F}; 
   	DqList = {qMap.F.DuList[[1]]};
   	DSList = {Table[Sum[SList[[j]]*DqList[[1,j]],{j,1,i}],{i,1,n}]}; 
   	For[k = 1, k <= Degree, k++,
   		DSJointk = {};
    		For[i = 1, i <= n, i++,
     			tmp = Sum[Binomial[k - 1, l]*SE3ad[DSList[[l+SHIFT, i]],DSJointList[[k-l-1+SHIFT,i]]],{l,0,k-1}];
     			DSJointk = Append[DSJointk, tmp];
     		];
    		DSJointList = Append[DSJointList, DSJointk];
    		
    		tmp = Transpose[Table[DSJointList[[k+SHIFT, Iu[[j]]]], {j, 1, nu}]];
    		DJuk = Table[tmp[[P[[j]]]], {j, 1, Length[P]}];
    		DJuList = Append[DJuList, DJuk]; 
    
    		tmp = Transpose[Table[DSJointList[[k+SHIFT, Id[[j]]]], {j, 1, nd}]];
    		DJdk = Table[tmp[[P[[j]]]], {j, 1, Length[P]}];
    		DJdList = Append[DJdList, DJdk];
    	
    		DJdInvk = -JdInv.Sum[Binomial[k, l]*DJdList[[l+SHIFT]].DJdInvList[[k - l +SHIFT]],{l,1,k}];
    		DJdInvList = Append[DJdInvList, DJdInvk];
    		
    		tmp = -Sum[Binomial[k, l]*DJdInvList[[l+SHIFT]].DJuList[[k - l +SHIFT]],{l,0,k}];
    		DDList = Append[DDList, tmp]; 
    		DFList = Append[DFList, ArrayFlatten[{{tmp}, {0*IdentityMatrix[nu]}}]]; 
    
    		kq = k + 1;
    		tmp = Sum[Binomial[kq - 1, l]*DFList[[l+SHIFT]].DuList[[kq-l]],{l,0,kq-1}];
    		DqList = Append[DqList, qMap.tmp]; (* q-vector according to partitioning in dep. and ind. *) 
    		
    		DSk = {};
    		For[i = 1, i <= n, i++,
     			tmp = Sum[Sum[Binomial[k, l]*DSJointList[[l+SHIFT,j]]*DqList[[k - l + 1, j]],{l,0,k}],{j,1,i}];
     			DSk = Append[DSk, tmp];
     		];
    		DSList = Append[DSList, DSk];
    	];
    	Return[DqList];
   ];
Protect[Dqdt];

Unprotect[ZeroVector];
ZeroVector[n_] := Module[{i}, 
  Table[0,{n}]
]
Protect[ZeroVector];

Unprotect[SE3InvolutiveClosure];
SE3InvolutiveClosure[X_,maxIteration__:3] := Module[{XX},
	If[ Dimensions[X][[1]]<3, XX=Append[X,X[[1]]],XX=X];
	SE3ClosureBasis[XX,XX,1,maxIteration]
	];
Protect[SE3InvolutiveClosure];

(****	Determine a basis for the involutive closure of a distribution		****)
(****	spanned by a set of se(3) vector fields 				****)
(****   !!!! does NOT work for less then three X_i !!!!!!!!!!!!!!!!!            ****)
SE3ClosureBasis[X_, Y_, count_, maxIteration_] := 
	Module[{n = Dimensions[X][[1]], b, i, j,YY=Y}, 
	If[count <= maxIteration, 
            b = Flatten[Table[SE3ad[X[[i]], YY[[j]]], {i, 1, n - 1}, {j, i, n}], 1];
				Join[YY, SE3ClosureBasis[X, b, count + 1,maxIteration]],YY]
	];



(* Close up the open environments *)
End[];
EndPackage[];

(* End HigherOrderDerivatives.m *)
